__author__ = "Sukanya Kalluri <sukkallu@cisco.com>"

import json
import logging
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation
from rest.connector.utils import get_username_password

# create a logger for this module
log = logging.getLogger(__name__)
STD_HEADER = {'Content-type': 'application/json'}


class Implementation(Implementation):
    '''Rest Implementation for DCNM

    Implementation of Rest connection to devices based on pyATS BaseConnection
    for DCNM

    YAML Example
    ------------

        devices:
            dcnm:
                connections:
                    vty:
                        protocol : ssh
                        ip : "2.3.4.5"
                    rest:
                        class: rest.connector.Rest
                        ip : "2.3.4.5"
                        port: 443
                        protocol: https
                        credentials:
                            rest:
                                username: admin
                                password: cisco123

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/dcnm123.yaml')
        >>> device = testbed.devices['dcnm']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''

    @BaseConnection.locked
    def connect(self, timeout=10, port=443, protocol='https'):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value
            protocol (str): http or https. Default to https

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.


        YAML Example
        ------------

            devices:
                dcnm:
                    os: dcnm
                    connections:
                        vty:
                            protocol : ssh
                            ip : "2.3.4.5"
                            protocol: https
                        rest:
                            class: rest.connector.Rest
                            ip : "2.3.4.5"
                            port: "443"
                            protocol: https
                            credentials:
                                rest:
                                    username: admin
                                    password: cisco123

        Code Example
        ------------

            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/dcnm123.yaml')
            >>> device = testbed.devices['haPrimary']
            >>> device.connect(alias='rest', via='rest')
        '''

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    '`unicon` is not installed for `sshtunnel`. Please install by `pip install unicon`.'
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            # prefer host instead of ip address directly
            try:
                host = self.connection_info['host']
            except KeyError:
                host = self.connection_info['ip'].exploded
            port = self.connection_info.get('port', port)

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.url = '{protocol}://{host}:{port}'.format(protocol=protocol,
                                                          host=host,
                                                          port=port)

        self.verify = self.connection_info.get('verify', True)

        username, password = get_username_password(self)

        _data = json.dumps({'expirationTime': 999999})

        login_url = '{url}/rest/logon'.format(url=self.url)

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        self.session = requests.Session()
        self.session.auth = HTTPBasicAuth(username, password)
        response = self.session.post(login_url, data=json.dumps({'expirationTime': 999999}), headers = STD_HEADER, verify=self.verify)
        log.info(response)

        # Make sure it returned requests.codes.ok
        if response.status_code != requests.codes.ok:
            # Something bad happened
            raise RequestException("Connection to '{ip}' has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{ok}'" \
                                   .format(ip=host, c=response.status_code,
                                           ok=requests.codes.ok))

        self.token = response.json()['Dcnm-Token']  # Retrieve the Token from the returned JSONhahhah

        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))
        return self.token


    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    def isconnected(func):
        '''Decorator to make sure session to device is active

           There is limitation on the amount of time the session ca be active
           for on the DCNM devices. However, there are no way to verify if
           session is still active unless sending a command. So, its just
           faster to reconnect every time.
         '''
        def decorated(self, *args, **kwargs):
            # Check if connected
            try:
                log.propagate = False
                self.disconnect()
                self.connect()
                log.propagate = True
                print(kwargs)
                ret = func(self, *args, **kwargs)
            finally:
                log.propagate = True
            return ret
        return decorated

    @BaseConnection.locked
    def _request(self, method, api_url, **kwargs):
        """ Wrapper to send REST command to device

        Args:
            method (str): session request method

            dn (str): rest endpoint

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok

        """
        if not self.connected:
            raise Exception("'{d}' is not connected for alias '{a}'"
                            .format(d=self.device.name,
                                    a=self.alias))

        # Deal with the dn
        full_url = '{f}{api_url}'.format(f=self.url, api_url=api_url)

        if 'data' in kwargs:
            payload = kwargs['data']
        elif 'json' in kwargs:
            payload = kwargs['json']
        else:
            payload = ''

        log.info("Sending {method} command to '{d}':"
                 "\napi_url: {furl}\nPayload:{payload}"
                 .format(method=method,
                         d=self.device.name,
                         furl=full_url,
                         payload=payload))

        headers = {
            'Dcnm-Token': self.token,
            'Content-type': 'application/json'
        }

        # Send to the device
        response = self.session.request(method=method, url=full_url, headers=headers, **kwargs)


        # Make sure it was successful
        try:
            response.raise_for_status()
        except Exception:
            raise RequestException(
                "'{c}' result code has been returned "
                "for '{d}'.\nResponse from server: "
                "{r}".format(d=self.device.name,
                             c=response.status_code,
                             r=response.text))

        # In case the response cannot be decoded into json
        # warn and return the raw text

        try:
            output = response.json()
            if isinstance(output, list):
                output = {"temproot": output}
        except Exception:
            log.warning('Could not decode json. Returning text!')
            output = response.text

        return output

    @BaseConnection.locked
    @isconnected
    def get(self, api_url, timeout=30):
        """ GET REST Command to retrieve information from the device

        Args:
            dn (str): Unique distinguished name that describes the object
                      and its place in the tree.

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time to allow rest call to return

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        return self._request('GET', api_url, timeout=timeout)

    @BaseConnection.locked
    @isconnected
    def post(self, api_url, payload, timeout=30):
        """POST REST Command to configure new information on the device

        Args:
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            payload (dict): Dictionary containing the information to send via
                            the post

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        return self._request('POST', api_url, data=json.dumps(payload), timeout=timeout)

    @BaseConnection.locked
    @isconnected
    def delete(self, api_url, timeout=30):
        """DELETE REST Command to delete information from the device

        Args
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        return self._request('DELETE', api_url, timeout=timeout)

    @BaseConnection.locked
    @isconnected
    def patch(self, api_url, payload, timeout=30):
        """PATCH REST Command to partially update existing information on the device

        Args
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            payload (dict): Dictionary containing the information to send via
                            the post

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        return self._request('PATCH', api_url, data=json.dumps(payload),
                             timeout=timeout)

    @BaseConnection.locked
    @isconnected
    def put(self, api_url, payload, timeout=30):
        """PUT REST Command to update existing information on the device

        Args
            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.

            payload (dict): Dictionary containing the information to send via
                            the post

            headers (dict): Headers to send with the rest call

            timeout (int): Maximum time

        Returns:
            response.json() or response.text

        Raises:
            RequestException if response is not ok
        """
        return self._request('DELETE', api_url, data=json.dumps(payload),
                             timeout=timeout)



""" Utilities shared by all plugin libraries. """

try:
    from pyats.utils.secret_strings import to_plaintext
except ImportError:
    def to_plaintext(string):
        return(str(string))

def get_username_password(connection):
    username = password = None
    if connection.connection_info.get('credentials'):
        try:
            username = str(
                connection.connection_info['credentials']['rest']['username'])
            password = to_plaintext(
                connection.connection_info['credentials']['rest']['password'])
        except Exception:
            pass

    if not username:
        username = connection.connection_info.get('username', \
            connection.device.tacacs.username \
            if connection.device.tacacs.username \
            else 'admin')

    if not password:
        password = connection.connection_info.get('password', \
            connection.device.passwords.tacacs \
            if connection.device.passwords.tacacs \
            else 'admin')

    return (username, password)



__version__   = '21.2.1'
__author__    = 'Jean-Benoit Aubin <jeaubin@cisco.com>'
__date__      = ''
__contact__   = 'pyats-support@cisco.com'
__copyright__ = 'Cisco System, Inc. Cisco Confedential'

__import__('pkg_resources').declare_namespace(__name__)



import json
import logging
import requests

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation
from rest.connector.utils import get_username_password

from requests.exceptions import RequestException
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Implementation):
    '''Rest Implementation for viptela/vManage

    Implementation of Rest connection to devices based on pyATS BaseConnection
    for vManage

    YAML Example
    ------------

        devices:
            vmanage:
                os: viptela
                type: vmanage
                custom:
                    abstraction:
                        order: [os]
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip : "2.3.4.5"
                        port: "8443"
                        verify: False
                        protocol: https
                        credentials:
                            rest:
                                username: admin
                                password: cisco123

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/sdwan.yaml')
        >>> device = testbed.devices['vmanage']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''

    @BaseConnection.locked
    def connect(self, timeout=30, port="8443", protocol='https'):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value
            port (str): Port number. Default to 8443
            protocol (str): http or https. Default to https

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.
        '''

        if self.connected:
            return

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    '`unicon` is not installed for `sshtunnel`. Please install by `pip install unicon`.'
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            ip = self.connection_info['ip'].exploded
            port = self.connection_info.get('port', port)

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.base_url = '{protocol}://{ip}:{port}'.format(protocol=protocol,
                                                          ip=ip,
                                                          port=port)

        self.verify = self.connection_info.get('verify', False)

        username, password = get_username_password(self)


        login_action = '/j_security_check'
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        # Format data for loginForm
        login_data = {'j_username': username, 'j_password': password}

        # Url for posting login data
        login_url = self.base_url + login_action

        token_url = '{url}/dataservice/client/token'.format(url=self.base_url)

        self.session = requests.session()
        resp = self.session.post(login_url,
                                 data=login_data,
                                 headers=headers,
                                 verify=self.verify,
                                 timeout=timeout)

        if resp.status_code == 200:
            log.info("Login successfully to '{d}'".format(d=self.device.name))
        else:
            raise RequestException(
                "Failed to login '{d}'".format(d=self.device.name))

        login_token = self.session.get(url=token_url,
                                       verify=self.verify,
                                       timeout=timeout)

        if login_token.status_code == 200:
            self.token = login_token.content
        else:
            raise RequestException(
                "Failed to get token for '{d}'".format(d=self.device.name))

        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))
        self.default_headers = {'X-XSRF-TOKEN': self.token,
                                'Content-Type': 'application/json'}

    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def get(self,
            mount_point,
            headers=None,
            timeout=30):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            mount_point (string): API url string
            headers (dict): Additional headers dictionary
            timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        data_url = self.base_url
        full_url = '{url}/{mount_point}'.format(
            url=data_url,
            mount_point=mount_point)

        log.info("Sending GET command to '{d}':"
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        hdr = self.default_headers

        if headers is not None:
            hdr.update(headers)

        response = self.session.get(full_url, headers=hdr,
                                    verify=self.verify, timeout=timeout)
        log.info("Output received:\n{response}".format(response=response))

        return response

    @BaseConnection.locked
    def post(self,
             mount_point,
             payload,
             headers=None,
             timeout=30):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            mount_point (string): API url string
            payload (dict): payload dictionary
            headers (dict): Additional headers dictionary
            timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        data_url = self.base_url
        full_url = '{url}/{mount_point}'.format(
            url=data_url,
            mount_point=mount_point)

        log.info("Sending POST command to '{d}':"
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        if isinstance(payload, dict):
            payload = json.dumps(payload)

        hdr = self.default_headers

        if headers is not None:
            hdr.update(headers)

        response = self.session.post(full_url, data=payload, headers=hdr,
                                     verify=self.verify, timeout=timeout)
        log.info("Output received:\n{response}".format(response=response))

        return response

    @BaseConnection.locked
    def put(self,
            mount_point,
            payload,
            headers=None,
            timeout=30):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            mount_point (string): API url string
            payload (dict): payload dictionary
            headers (dict): Additional headers dictionary
            timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        data_url = self.base_url
        full_url = '{url}/{mount_point}'.format(
            url=data_url,
            mount_point=mount_point)

        log.info("Sending PUT command to '{d}':"
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        if isinstance(payload, dict):
            payload = json.dumps(payload)

        hdr = self.default_headers

        if headers is not None:
            hdr.update(headers)

        response = self.session.put(full_url, data=payload, headers=hdr,
                                    verify=self.verify, timeout=timeout)
        log.info("Output received:\n{response}".format(response=response))

        return response

    @BaseConnection.locked
    def delete(self,
               mount_point,
               headers=None,
               timeout=30):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            mount_point (string): API url string
            payload (dict): payload dictionary
            headers (dict): Additional headers dictionary
            timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        data_url = self.base_url
        full_url = '{url}/{mount_point}'.format(
            url=data_url,
            mount_point=mount_point)

        log.info("Sending DELETE command to '{d}':"
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        hdr = self.default_headers

        if headers is not None:
            hdr.update(headers)

        response = self.session.delete(full_url, headers=hdr,
                                       verify=self.verify, timeout=timeout)
        log.info("Output received:\n{response}".format(response=response))

        return response



# Global imports
import logging
import time

# Genie, pyATS, ROBOT imports
# from pyats.connections import BaseConnection
from rest.connector.utils import get_username_password
from rest.connector.implementation import Implementation
from pyats.connections import BaseConnection

# F5 imports
from icontrol.session import iControlRESTSession
from icontrol.exceptions import iControlUnexpectedHTTPError

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Implementation):

    """Rest BaseClass



    Baseclass for Rest connection implementation



    YAML Example

    ------------



        devices:

            bigip1:
                type: 'generic'
                os: 'bigip'
                custom:
                    abstraction:
                        order: [os]
                connections:
                    # Console
                    a:
                        ip: 1.2.3.4
                        port: 22
                        protocol: ssh
                    rest:
                        # specify the rest connector class
                        class: rest.connector.Rest
                        ip: 1.2.3.4
                        port: "443"
                        protocol: https
                        credentials:
                            rest:
                                username: user
                                password: password



    Example

    -------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/path/to/testbed.yaml')
        >>> device = testbed.devices['bigip1']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected

        True

    """

    @property
    def connected(self):

        """Is a device connected"""

        return self._is_connected

    def disconnect(self):

        """disconnect the device for this particular alias"""

        try:
            log.info("Deleting token: '{t}'".format(t=self.token))
            delete_url = "https://{0}:{1}/mgmt/shared/authz/tokens/{2}".format(
                self.ip, self.port, self.token
            )

            delete_icr_session = iControlRESTSession(
                self.username,
                self.password,
                verify=self.verify,
                token_to_use=self.token
            )
            # Extending the timeout for the token received
            delete_icr_session.delete(delete_url)

            log.info("Token Deleted")

        finally:

            self._is_connected = False

        log.info(
            "'{t}' token deleted successfully.".format(t=self.token)
        )

    def isconnected(func):

        '''Decorator to make sure session to device is active



           There is limitation on the amount of time the session ca be active

           for on the NXOS devices. However, there are no way to verify if

           session is still active unless sending a command. So, its just

           faster to reconnect every time.

         '''

        def decorated(self, *args, **kwargs):

            # Check if connected

            try:

                log.propagate = False

                self.disconnect()

                if 'timeout' in kwargs:
                    self.connect(timeout=kwargs['timeout'])
                else:
                    self.connect()

                log.propagate = True

                ret = func(self, *args, **kwargs)

            finally:

                log.propagate = True

            return ret

        return decorated

    def connect(self, auth_provider='tmos', verify=False, port='443', protocol='https', *args, **kwargs):

        if self.connected:
            return

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    '`unicon` is not installed for `sshtunnel`. Please install by `pip install unicon`.'
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            ip = self.connection_info['ip'].exploded
            port = self.connection_info.get('port', port)

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.base_url = '{protocol}://{ip}:{port}'.format(protocol=protocol,
                                                          ip=ip,
                                                          port=port)

        self.username, self.password = get_username_password(self)

        self.header = "Content-Type: application/json"
        self.verify = verify

        # URL to authenticate and receive the token
        url = "https://{0}:{1}/mgmt/shared/authn/login".format(
            self.ip, self.port
        )
        payload = {
            'username': self.username,
            'password': self.password,
            'loginProviderName': auth_provider
        }

        iCRS = iControlRESTSession(
            self.username,
            self.password,
            verify=self.verify
        )

        log.info(
            "Connecting to '{d}' with alias "
            "'{a}'".format(d=self.device.name, a=self.alias)
        )

        response = iCRS.post(
            url,
            json=payload,
        )

        log.info(response.json())

        if response.status_code not in [200]:
            if b'Configuration Utility restarting...' in response.content:
                time.sleep(30)
                # self.retries += 1
                return self.connect()
            else:
                # self.retries = 0
                return None, response.content

        self.token = response.json()['token']['token']

        log.info("The following toke is used to connect'{t}'".format(t=self.token))

        # Self-link of the token
        timeout_url = "https://{0}:{1}/mgmt/shared/authz/tokens/{2}".format(
            self.ip, self.port, self.token
        )
        timeout_payload = {"timeout": "3600"}

        token_icr_session = iControlRESTSession(
            self.username,
            self.password,
            verify=self.verify,
            token_to_use=self.token
        )

        # Extending the timeout for the token received
        token_icr_session.patch(timeout_url, json=timeout_payload)

        log.info("'{t}' - Token timeout extended to '{time}'".format(t=self.token, time=timeout_payload))

        params = dict(
            username=self.username,
            password=self.password,
            verify=self.verify,
            token_to_use=self.token
        )

        # creating an object to be used all new requests
        self.icr_session = iControlRESTSession(**params)

        self._is_connected = True

        log.info("Connected successfully to '{d}' using token: '{t}'".format(d=self.device.name, t=self.token))

        return self._is_connected, self.icr_session

    @isconnected
    def get(self, api_url, timeout=30, verbose=False):

        """GET REST Command to retrieve information from the device"""

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending GET to '{d}': "
            "{u}".format(d=self.device.name, u=full_url)
        )

        response = self.icr_session.get(full_url, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        log.info("Successfully fetched data using token: '{t}'".format(t=self.token))

        return output

    @BaseConnection.locked
    @isconnected
    def post(self, api_url, payload, timeout=30, verbose=False):
        """POST REST Command to configure information from the device"""

        if not self.connected:
            raise Exception(
                "'{d}' is not connected for "
                "alias '{a}'".format(d=self.device.name, a=self.alias)
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'"
            " and payload '{p}'".format(
                d=self.device.name, u=full_url, h=self.header, p=payload
            )
        )

        response = self.icr_session.post(full_url, json=payload, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    @BaseConnection.locked
    @isconnected
    def put(self, api_url, payload, timeout=30, verbose=False):

        """PUT REST Command to update information on the device"""

        if not self.connected:
            raise Exception(
                "'{d}' is not connected for "
                "alias '{a}'".format(d=self.device.name, a=self.alias)
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'"
            " and payload '{p}'".format(
                d=self.device.name, u=full_url, h=self.header, p=payload
            )
        )

        response = self.icr_session.put(full_url, json=payload, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    @BaseConnection.locked
    @isconnected
    def patch(self, api_url, payload, timeout=30, verbose=False):

        """PATCH REST Command to update information on the device"""

        if not self.connected:
            raise Exception(
                "'{d}' is not connected for "
                "alias '{a}'".format(d=self.device.name, a=self.alias)
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'"
            " and payload '{p}'".format(
                d=self.device.name, u=full_url, h=self.header, p=payload
            )
        )

        response = self.icr_session.patch(full_url, json=payload, timeout=timeout)

        output = response

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    @BaseConnection.locked
    @isconnected
    def delete(self, api_url, timeout=30, verbose=False):

        """DELETE REST Command to delete information from the device"""

        if not self.connected:
            raise Exception(
                "'{d}' is not connected for "
                "alias '{a}'".format(d=self.device.name, a=self.alias)
            )

        full_url = "{b}{a}".format(b=self.base_url, a=api_url)

        log.info(
            "Sending Post to '{d}': "
            "{u}"
            " with the header '{h}'".format(
                d=self.device.name, u=full_url, h=self.header
            )
        )

        response = self.icr_session.delete(full_url, timeout=timeout)

        output = response.text

        log.debug(
            "Response: {c}, headers: {h}".format(
                c=response.status_code, h=self.header
            )
        )
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned ok
        if not response.ok:
            raise iControlUnexpectedHTTPError(
                "Connection to '{d}' has returned the "
                "following code '{c}', instead of the "
                "expected status code 'ok'".format(
                    d=self.device.name, c=response.status_code
                )
            )

        log.info(
            "Successfully fetched data from '{d}'".format(d=self.device.name)
        )

        return output

    def configure(self, *args, **kwargs):

        """configure - Not implemented for REST"""

        raise NotImplementedError(
            "configure is not a supported method for REST. "
            "post is probably what you are looking for"
        )

    def execute(self, *args, **kwargs):

        """execute - Not implemented for REST"""

        raise NotImplementedError(
            "execute is not a supported method for REST. "
            "get is probably what you are looking for."
        )



import re
import json
import logging
import requests
from dicttoxml import dicttoxml
from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation as RestImplementation
from rest.connector.utils import get_username_password


# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(RestImplementation):
    '''Rest Implementation for NSO

    Implementation of Rest connection to Network Service Orchestrator (NSO)

    YAML Example
    ------------

        devices:
            ncs:
                os: nso
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip: 127.0.0.1
                        port: "8080"
                        protocol: http
                        credentials:
                            rest:
                                username: admin
                                password: admin
                custom:
                    abstraction:
                        order: [os]

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('ncs.yaml')
        >>> device = testbed.devices['ncs']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''

    @BaseConnection.locked
    def connect(self,
                timeout=30,
                port='8080',
                protocol='http',
                default_content_type='json',
                verbose=False):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value
            port (str): Port number. Default to 8080
            protocol (str): http or https. Default to http
            default_content_type: Default for content type, json or xml

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.

        '''

        if self.connected:
            return

        log.debug("Content type: %s" % default_content_type)
        log.debug("Timeout: %s" % timeout)
        self.content_type = default_content_type

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    '`unicon` is not installed for `sshtunnel`. Please install by `pip install unicon`.'
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            ip = self.connection_info['ip'].exploded
            port = self.connection_info.get('port', port)

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.base_url = '{protocol}://{ip}:{port}'.format(protocol=protocol,
                                                          ip=ip,
                                                          port=port)

        self.login_url = '{f}/api'.format(f=self.base_url)

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        username, password = get_username_password(self)

        self.session = requests.Session()
        self.session.auth = (username, password)

        # Connect to the device via requests
        response = self.session.get(self.login_url, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
            r=response.reason, h=response.headers))
        if verbose:
            log.info("Response text:\n%s" % output)

        # Make sure it returned requests.codes.ok
        if response.status_code != requests.codes.ok:
            # Something bad happened
            raise RequestException("Connection to '{ip}:{port}' has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{ok}'"\
                                        .format(ip=ip, port=port, c=response.status_code,
                                                ok=requests.codes.ok))
        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

        return response


    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))


    @BaseConnection.locked
    def get(self, api_url, content_type=None, headers=None,
             expected_status_codes=(
             requests.codes.no_content,
             requests.codes.ok
             ),
            timeout=30,
            verbose=False):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------
        api_url: API url string
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        if content_type is None:
            content_type = self.content_type

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        header = 'application/vnd.yang.data+{fmt}' \
                 ', application/vnd.yang.collection+{fmt}' \
                 ', application/vnd.yang.datastore+{fmt}'

        if content_type.lower() == 'json':
            accept_header = header.format(fmt='json')
        elif content_type.lower() == 'xml':
            accept_header = header.format(fmt='xml')
        else:
            accept_header = content_type

        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending GET command to '{d}': "\
                 "{u}".format(d=self.device.name, u=full_url))
        log.debug("Request headers:{headers}".format(
                    headers= self.session.headers))

        response = self.session.get(full_url, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
            r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response


    @BaseConnection.locked
    def post(self, api_url, payload='', content_type=None, headers=None, 
             expected_status_codes=(
             requests.codes.created,
             requests.codes.no_content,
             requests.codes.ok
             ),
             timeout=30,
             verbose=False):
        '''POST REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        payload: payload to sent, can be string or dict
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        request_payload = payload
        if isinstance(payload, dict):
            assert content_type != None, 'content_type parameter required when passing dict'
            if content_type == 'json':
                request_payload = json.dumps(payload)
            elif content_type == 'xml':
                request_payload = dicttoxml(payload, attr_type=False)

        if content_type is None:
            if re.match("<", payload.lstrip()) is not None:
                content_type = 'xml'
            else:
                content_type = 'json'

        # NSO Northbound APIs / Resources / Operations and Actions
        #
        # YANG-defined operations, defined with the YANG statements "rpc" or "tailf:action",
        # and the built-in operations, are represented with the media type "application/vnd.yang.operation".
        # Resources of this type accept only the method "POST".
        # In XML, such resources are encoded as subelements to the XML element "y:operations".
        # In JSON, they are encoded under "_operations".
        #

        if '/_operations' in api_url:
            # Rest operations like /api/running/devices/_operations/connect
            content_type_header = 'application/vnd.yang.operation+{fmt}'
        elif '/api/' in api_url:
            # for URIs accessing datastores (i.e. /api/running))
            content_type_header = 'application/vnd.yang.datastore+{fmt}'
        else:
            content_type_header = 'application/vnd.yang.data+{fmt}'

        accept_header = 'application/vnd.yang.data+{fmt}' \
         ', application/vnd.yang.collection+{fmt}' \
         ', application/vnd.yang.datastore+{fmt}'

        if content_type.lower() == 'json':
            content_type_header = content_type_header.format(fmt='json')
            accept_header = accept_header.format(fmt='json')
        elif content_type.lower() == 'xml':
            content_type_header = content_type_header.format(fmt='xml')
            accept_header = accept_header.format(fmt='xml')
        else:
            content_type_header = content_type
            accept_header = content_type

        self.session.headers.update({'Content-type': content_type_header})
        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending POST command to '{d}': {u}"\
            .format(d=self.device.name, u=full_url))
        log.debug("Request headers: {h}\nPayload: {p}"\
            .format(h=self.session.headers, p=request_payload))
        if verbose:
            log.info('Request payload:\n{payload}'.format(payload=request_payload))

        # Send to the device
        response = self.session.post(full_url, request_payload, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
            r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response


    @BaseConnection.locked
    def patch(self, api_url, payload, content_type=None, headers=None, 
             expected_status_codes=(
             requests.codes.created,
             requests.codes.no_content,
             requests.codes.ok
             ),
             timeout=30,
             verbose=False):
        '''PATCH REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        payload: payload to sent, can be string or dict
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        request_payload = payload
        if isinstance(payload, dict):
            assert content_type != None, 'content_type parameter required when passing dict'
            if content_type == 'json':
                request_payload = json.dumps(payload)
            elif content_type == 'xml':
                request_payload = dicttoxml(payload, attr_type=False)

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        if content_type is None:
            if re.match("<", payload.lstrip()) is not None:
                content_type = 'xml'
            else:
                content_type = 'json'

        content_type_header = 'application/vnd.yang.data+{fmt}'
        accept_header = 'application/vnd.yang.data+{fmt}' \
         ', application/vnd.yang.collection+{fmt}' \
         ', application/vnd.yang.datastore+{fmt}'

        if content_type.lower() == 'json':
            content_type_header = content_type_header.format(fmt='json')
            accept_header = accept_header.format(fmt='json')
        elif content_type.lower() == 'xml':
            content_type_header = content_type_header.format(fmt='xml')
            accept_header = accept_header.format(fmt='xml')
        else:
            content_type_header = content_type
            accept_header = content_type

        self.session.headers.update({'Content-type': content_type_header})
        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending PATCH command to '{d}': {u}".format(
                                                    d=self.device.name, u=full_url))
        log.debug("Request headers: {h}\nPayload:{p}".format(h=self.session.headers,
                                                    p=request_payload))
        if verbose:
            log.info('Request payload:\n{payload}'.format(payload=request_payload))

        # Send to the device
        response = self.session.patch(full_url, request_payload, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
            r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response


    @BaseConnection.locked
    def put(self, api_url, payload, content_type=None, headers=None, 
             expected_status_codes=(
             requests.codes.created,
             requests.codes.no_content,
             requests.codes.ok
             ),
             timeout=30,
             verbose=False):
        '''PUT REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        payload: payload to sent, can be string or dict
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        request_payload = payload
        if isinstance(payload, dict):
            assert content_type != None, 'content_type parameter required when passing dict'
            if content_type == 'json':
                request_payload = json.dumps(payload)
            elif content_type == 'xml':
                request_payload = dicttoxml(payload, attr_type=False)

        if content_type is None:
            if re.match("<", payload.lstrip()) is not None:
                content_type = 'xml'
            else:
                content_type = 'json'

        content_type_header = 'application/vnd.yang.data+{fmt}'
        accept_header = 'application/vnd.yang.data+{fmt}' \
         ', application/vnd.yang.collection+{fmt}' \
         ', application/vnd.yang.datastore+{fmt}'

        if content_type.lower() == 'json':
            content_type_header = content_type_header.format(fmt='json')
            accept_header = accept_header.format(fmt='json')
        elif content_type.lower() == 'xml':
            content_type_header = content_type_header.format(fmt='xml')
            accept_header = accept_header.format(fmt='xml')
        else:
            content_type_header = content_type
            accept_header = content_type

        self.session.headers.update({'Content-type': content_type_header})
        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending PUT command to '{d}': {u}".format(
                                                    d=self.device.name, u=full_url))
        log.debug("Request headers: {h}\nPayload:{p}".format(h=self.session.headers,
                                                    p=request_payload))
        if verbose:
            log.info('Request payload:\n{payload}'.format(payload=request_payload))

        # Send to the device
        response = self.session.put(full_url, request_payload, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
            r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response


    @BaseConnection.locked
    def delete(self, api_url, content_type=None, headers=None, 
             expected_status_codes=(
             requests.codes.created,
             requests.codes.no_content,
             requests.codes.ok
             ),
             timeout=30,
             verbose=False):
        '''DELETE REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        if content_type is None:
            content_type = self.content_type

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        if content_type.lower() == 'json':
            accept_header = 'application/vnd.yang.data+json'
        elif content_type.lower() == 'xml':
            accept_header = 'application/vnd.yang.data+xml'
        else:
            accept_header = content_type

        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending DELETE command to '{d}': "\
                 "{u}".format(d=self.device.name, u=full_url))
        log.debug("Request headers:{headers}".format(
                    headers= self.session.headers))

        response = self.session.delete(full_url, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
            r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response



from pyats.connections import BaseConnection


class Implementation(BaseConnection):
    '''Rest BaseClass

    Baseclass for Rest connection implementation

    YAML Example
    ------------

        devices:
            PE1:
                credentials:
                    rest:
                        username: admin
                        password: cisco123
                connections:
                    a:
                        protocol: telnet
                        ip: "1.2.3.4"
                        port: 2004
                    vty:
                        protocol : telnet
                        ip : "2.3.4.5"
                    rest:
                        class: rest.connector.Rest
                        ip : "2.3.4.5"
                        port: "443"
                        protocol: https
                        credentials:
                            rest:
                                username: admin
                                password: cisco123

    Example
    -------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/asr22.yaml')
        >>> device = testbed.devices['PE1']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''

    def __init__(self, *args, **kwargs):
        '''__init__ instantiates a single connection instance.'''

        # instanciate BaseConnection
        # (could use super...)
        BaseConnection.__init__(self, *args, **kwargs)
        self._is_connected = False

    @property
    def connected(self):
        '''Is a device connected'''
        return self._is_connected

    def connect(self, *args, **kwargs):
        '''connect to the device via REST'''

        raise NotImplementedError

    def disconnect(self):
        '''disconnect the device for this particular alias'''

        raise NotImplementedError

    def get(self, *args, **kwargs):
        '''GET REST Command to retrieve information from the device'''

        raise NotImplementedError

    def post(self, *args, **kwargs):
        '''POST REST Command to configure information from the device'''

        raise NotImplementedError

    def put(self, *args, **kwargs):
        '''PUT REST Command to update information on the device'''

        raise NotImplementedError

    def patch(self, *args, **kwargs):
        '''PATCH REST Command to update information on the device'''

        raise NotImplementedError

    def delete(self, *args, **kwargs):
        '''DELETE REST Command to delete information from the device'''

        raise NotImplementedError

    def configure(self, *args, **kwargs):
        '''configure - Not implemented for REST'''
        raise NotImplementedError('configure is not a supported method for REST. '
                                  'post is probably what you are looking for')

    def execute(self, *args, **kwargs):
        '''execute - Not implemented for REST'''
        raise NotImplementedError('execute is not a supported method for REST. '
                                  'get is probably what you are looking for.')



import json
import logging
import requests

from requests.exceptions import RequestException


from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation as Imp
from rest.connector.utils import get_username_password

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Imp):
    '''Rest Implementation for APIC

    Implementation of Rest connection to devices based on pyATS BaseConnection
    for APIC

    YAML Example
    ------------

        devices:
            apic1:
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip : "2.3.4.5"
                        credentials:
                            rest:
                                username: admin
                                password: cisco123

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
        >>> device = testbed.devices['apic1']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''
    def __init__(self, *args, **kwargs):
        import warnings
        warnings.warn(
            "This rest.connector library is deprecated and will be removed "
            "on v20.3. Please set your testbed to 'os: apic' in order to "
            "use the new library.")

        super().__init__(*args, **kwargs)

    @BaseConnection.locked
    def connect(self, timeout=30):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.


        YAML Example
        ------------

            devices:
                apic1:
                    connections:
                        rest:
                            class: rest.connector.Rest
                            ip : "2.3.4.5"
                            credentials:
                                rest:
                                    username: admin
                                    password: cisco123

        Code Example
        ------------

            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
            >>> device = testbed.devices['apic1']
            >>> device.connect(alias='rest', via='rest')
        '''

        if self.connected:
            return

        if 'host' in self.connection_info:
            ip = self.connection_info['host']
        else:
            ip = self.connection_info['ip'].exploded
        if 'port' in self.connection_info:
            port = self.connection_info['port']
            self.url = 'https://{ip}:{port}/'.format(ip=ip, port=port)
        else:
            self.url = 'https://{ip}/'.format(ip=ip)
        login_url = '{f}api/aaaLogin.json'.format(f=self.url)

        username, password = get_username_password(self)

        payload = {
           "aaaUser": {
              "attributes": {
                 "name": username,
                 "pwd": password,
               }
           }
        }
        headers = {
            'Content-Type': 'text/plain'
        }

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        self.session = requests.Session()
        _data = json.dumps(payload)

        # Connect to the device via requests
        response = self.session.post(login_url, data=_data, timeout=timeout, \
            verify=False, headers=headers)
        log.info(response)

        # Make sure it returned requests.codes.ok
        if response.status_code != requests.codes.ok:
            # Something bad happened
            raise RequestException("Connection to '{ip}' has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{ok}'"\
                                        .format(ip=ip, c=response.status_code,
                                                ok=requests.codes.ok))
        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    def isconnected(func):
        '''Decorator to make sure session to device is active

           There is limitation on the amount of time the session cab be active
           on the APIC. However, there are no way to verify if
           session is still active unless sending a command. So, its just
           faster to reconnect every time.
         '''
        def decorated(self, *args, **kwargs):
            # Check if connected
            try:
                self.disconnect()

                if 'timeout' in kwargs:
                    self.connect(timeout=kwargs['timeout'])
                else:
                    self.connect()
            finally:
                ret = func(self, *args, **kwargs)
            return ret
        return decorated

    @BaseConnection.locked
    @isconnected
    def get(self, dn, query_target='self', rsp_subtree='no', \
            query_target_filter='', rsp_prop_include='all', \
            rsp_subtree_include='', rsp_subtree_class='',\
            expected_status_code=requests.codes.ok, timeout=30):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.
            query_target {self|children|subtree}: 
                                'self': (default) MO itself
                                'children': just the MO's child objects
                                'subtree': MO and its child objects
            rsp_subtree {no|children|full}: Specifies child object level 
                                            included in the response
                                            'no': (default) the response
                                                   does not include any children
                                            'children': return only the child 
                                                        objects
                                            'full': includes the full tree 
                                                    structure
            rsp_prop_include {all|naming-only|config-only}:
                                'all': all properties of the objects
                                'naming-only': only the naming properties
                                'config-only': only configurable properties
            rsp_subtree_include (string): specify additional contained objects 
                                          or options to be included
            rsp_subtree_class (string) : specify classes
            query_target_filter (string): filter expression
            expected_status_code (int): Expected result
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = "{f}{dn}?query-target={qt}&rsp-subtree={rs}"\
                        "&rsp-prop-include={rpi}"\
                          .format(f=self.url,
                                  dn=dn,
                                  qt=query_target,
                                  rs=rsp_subtree,
                                  rpi=rsp_prop_include)
        if query_target_filter:
            full_url += "&query-target-filter={qtf}"\
                .format(qtf=query_target_filter)

        if rsp_subtree_include:
            full_url += "&rsp-subtree-include={rsi}"\
                .format(rsi=rsp_subtree_include)

        if rsp_subtree_class:
            full_url += "&rsp-subtree-class={rsc}"\
                .format(rsc=rsp_subtree_class)

        log.info("Sending GET command to '{d}':"\
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        response = self.session.get(full_url, timeout=timeout, verify=False)
        
        try:
            output = response.json()
        except Exception:
            output = response.text

        log.info("Output received:\n{output}".format(output=
            json.dumps(output, indent=2, sort_keys=True)))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("Sending '{furl} to '{d} has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{e}'"
                                   "'{e}'".format(furl=full_url,
                                                  d=self.device.name,
                                                  c=response.status_code,
                                                  e=expected_status_code))
        return output

    @BaseConnection.locked
    @isconnected
    def post(self, dn, payload, expected_status_code=requests.codes.ok,
             timeout=30):
        '''POST REST Command to configure information from the device

        Arguments
        ---------

            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.
            payload (dict): Dictionary containing the information to send via
                            the post
            expected_status_code (int): Expected result
            timeout (int): Maximum time
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        # Deal with the dn
        full_url = '{f}{dn}'.format(f=self.url, dn=dn)

        log.info("Sending POST command to '{d}':"\
                 "\nDN: {furl}\nPayload:{p}".format(d=self.device.name,
                                                    furl=full_url,
                                                    p=payload))

        # Send to the device
        response = self.session.post(full_url, payload, timeout=timeout, \
            verify=False)
        output = response.json()
        log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}', got:\n {msg}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code,
                                           msg=response.text))
        return output

    @BaseConnection.locked
    @isconnected
    def delete(self, dn, expected_status_code=requests.codes.ok, timeout=30):
        '''DELETE REST Command to delete information from the device

        Arguments
        ---------

            dn (string): Unique distinguished name that describes the object
                         and its place in the tree.
            expected_status_code (int): Expected result
            timeout (int): Maximum time
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        # Deal with the dn
        full_url = '{f}{dn}'.format(f=self.url, dn=dn)

        log.info("Sending DELETE command to '{d}':"\
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        # Send to the device
        response = self.session.delete(full_url, timeout=timeout, verify=False)
        output = response.json()
        log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}'"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code))
        return output



  
# Enable abstraction using this directory name as the abstraction token
try:
    from genie import abstract
    abstract.declare_token(__name__)
except Exception as e:
    import warnings
    warnings.warn('Could not declare abstraction token: ' + str(e))



#!/bin/env python
""" Unit tests for F5 (BigIP) rest.connector """

import os
import unittest
import requests
import requests_mock
from requests.models import Response
from unittest import mock
from unittest.mock import patch, Mock
from requests.exceptions import RequestException

from pyats.topology import loader

from rest.connector import Rest
from icontrol.session import iControlRESTSession

HERE = os.path.dirname(__file__)


class FakeResponse(object):
    status_code = 200

    def json(self):
        return {
            "username": "admin",
            "loginReference": {
                "link": "https://localhost/mgmt/cm/system/authn/providers/tmos/1f44a60e-11a7-3c51-a49f-82983026b41b/login"
            },
            "loginProviderName": "tmos",
            "token": {
                "token": "3UCPWZW66ZHOR6BUMVW56F6Q6K",
                "name": "3UCPWZW66ZHOR6BUMVW56F6Q6K",
                "userName": "admin",
                "authProviderName": "tmos",
                "user": {"link": "https://localhost/mgmt/shared/authz/users/admin"},
                "timeout": 1200,
                "startTime": "2019-12-03T05:25:10.321-0800",
                "address": "1.2.3.4",
                "partition": "[All]",
                "generation": 1,
                "lastUpdateMicros": 1575379510321149,
                "expirationMicros": 1575380710321000,
                "kind": "shared:authz:tokens:authtokenitemstate",
                "selfLink": "https://localhost/mgmt/shared/authz/tokens/3UCPWZW66ZHOR6BUMVW56F6Q6K",
            },
            "generation": 0,
            "lastUpdateMicros": 0,
        }


class FakeResponseTimout(object):
    status_code = 200

    def json(self):
        return {
            "token": {"token": "3UCPWZW66ZHOR6BUMVW56F6Q6K", "timeout": 3600,},
        }


class FakeResponseGet(object):
    status_code = 200

    def json(self):
        return {
            "kind": "tm:ltm:global-settings:global-settingscollectionstate",
            "selfLink": "https://localhost/mgmt/tm/ltm/global-settings?ver=13.0.1",
        }


class FakeResponsePost(object):
    status_code = 200

    def json(self):
        return {
            "kind": "tm:ltm:node:nodestate",
            "name": "wa122",
            "partition": "Common",
            "fullPath": "/Common/wa122",
            "generation": 2139,
            "selfLink": "https://localhost/mgmt/tm/ltm/node/~Common~wa122?ver=13.0.1",
            "address": "119.119.192.194",
            "connectionLimit": 0,
            "dynamicRatio": 1,
            "ephemeral": "false",
            "fqdn": {
                "addressFamily": "ipv4",
                "autopopulate": "disabled",
                "downInterval": 5,
                "interval": "3600",
            },
            "logging": "disabled",
            "monitor": "default",
            "rateLimit": "disabled",
            "ratio": 1,
            "session": "user-enabled",
            "state": "unchecked",
        }


class FakeResponsePatch(object):
    status_code = 200

    def json(self):
        return {
            "kind": "tm:ltm:node:nodestate",
            "name": "wa122",
            "partition": "Common",
            "fullPath": "/Common/wa122",
            "generation": 2140,
            "selfLink": "https://localhost/mgmt/tm/ltm/node/~Common~wa122?ver=13.0.1",
            "address": "119.119.192.194",
            "connectionLimit": 0,
            "dynamicRatio": 1,
            "ephemeral": "false",
            "fqdn": {
                "addressFamily": "ipv4",
                "autopopulate": "disabled",
                "downInterval": 5,
                "interval": "3600",
            },
            "logging": "disabled",
            "monitor": "default",
            "rateLimit": "disabled",
            "ratio": 1,
            "session": "user-disabled",
            "state": "unchecked",
        }


class FakeResponsePut(object):
    status_code = 200

    def json(self):
        return {
            "kind": "tm:ltm:pool:poolstate",
            "name": "wa12",
            "fullPath": "wa12",
            "generation": 2142,
            "selfLink": "https://localhost/mgmt/tm/ltm/pool/wa12?ver=13.0.1",
            "allowNat": "yes",
            "allowSnat": "yes",
            "ignorePersistedWeight": "disabled",
            "ipTosToClient": "pass-through",
            "ipTosToServer": "pass-through",
            "linkQosToClient": "pass-through",
            "linkQosToServer": "pass-through",
            "loadBalancingMode": "round-robin",
            "minActiveMembers": 0,
            "minUpMembers": 0,
            "minUpMembersAction": "failover",
            "minUpMembersChecking": "disabled",
            "queueDepthLimit": 0,
            "queueOnConnectionLimit": "disabled",
            "queueTimeLimit": 0,
            "reselectTries": 0,
            "serviceDownAction": "none",
            "slowRampTime": 10,
            "membersReference": {
                "link": "https://localhost/mgmt/tm/ltm/pool/~Common~wa12/members?ver=13.0.1",
                "isSubcollection": True,
            },
        }


class test_rest_connector(unittest.TestCase):
    def setUp(self):
        self.testbed = loader.load(os.path.join(HERE, "testbed.yaml"))
        self.device = self.testbed.devices["bigip01"]
        self.ip = str(self.device.connections.rest.ip)
        self.username, self.password = "admin", "admin"
        self.url = "https://" + str(self.ip) + "/mgmt/shared/authn/login"
        self.payload = {
            "username": self.username,
            "password": self.password,
            "loginProviderName": "tmos",
        }

    def test_init(self):
        connection = Rest(device=self.device, alias="rest", via="rest")
        self.assertEqual(connection.device, self.device)

        with self.assertRaises(NotImplementedError):
            self.assertRaises(connection.execute())
        with self.assertRaises(NotImplementedError):
            self.assertRaises(connection.configure())

    @patch("icontrol.session.iControlRESTSession.post")
    def test_iCRS(self, mock_connect_post):
        iCRS = iControlRESTSession(self.username, self.password)
        mock_connect_post.return_value.status_code = 200
        mock_connect_post.return_value = FakeResponse()
        response = iCRS.post(self.url, json=self.payload,)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json()["token"]["token"], "3UCPWZW66ZHOR6BUMVW56F6Q6K"
        )
        self.assertEqual(response.json()["token"]["timeout"], 1200)

    @patch("icontrol.session.iControlRESTSession.patch")
    def test_iCRS_token(self, mock_connect_patch):
        token = FakeResponse().json()["token"]["token"]

        timeout_payload = {"timeout": "3600"}
        timeout_url = "https://" + self.ip + "/mgmt/shared/authz/tokens/token"

        mock_connect_patch.return_value.status_code = 200
        mock_connect_patch.return_value = FakeResponseTimout()

        token_icr_session = iControlRESTSession(
            self.username, self.password, token_to_use=token
        )
        t_timeout = token_icr_session.patch(timeout_url, json=timeout_payload)

        self.assertEqual(
            t_timeout.json()["token"]["token"], "3UCPWZW66ZHOR6BUMVW56F6Q6K"
        )
        self.assertEqual(t_timeout.json()["token"]["timeout"], 3600)

    @patch("icontrol.session.iControlRESTSession.get")
    def test_get(self, mock_connect_get):
        full_url = "https://" + self.ip + "/mgmt/tm/ltm/global-settings"
        token = "3UCPWZW66ZHOR6BUMVW56F6Q6K"

        mock_connect_get.return_value.status_code = 200
        mock_connect_get.return_value = FakeResponseGet()

        icr_session = iControlRESTSession(
            self.username, self.password, token_to_use=token
        )
        get_icr_session = icr_session.get(full_url)

        output = {
            "kind": "tm:ltm:global-settings:global-settingscollectionstate",
            "selfLink": "https://localhost/mgmt/tm/ltm/global-settings?ver=13.0.1",
        }

        self.assertEqual(get_icr_session.json(), output)

    @patch("icontrol.session.iControlRESTSession.post")
    def test_post(self, mock_connect_post):
        full_url = "https://" + self.ip + "/mgmt/tm/ltm/node/"
        token = "3UCPWZW66ZHOR6BUMVW56F6Q6K"

        mock_connect_post.return_value.status_code = 200
        mock_connect_post.return_value = FakeResponsePost()

        icr_session = iControlRESTSession(
            self.username, self.password, token_to_use=token
        )

        data = {"name": "wa122", "partition": "Common", "address": "119.119.192.194"}

        post_icr_session = icr_session.post(full_url, json=data)

        output = {
            "kind": "tm:ltm:node:nodestate",
            "name": "wa122",
            "partition": "Common",
            "fullPath": "/Common/wa122",
            "generation": 2139,
            "selfLink": "https://localhost/mgmt/tm/ltm/node/~Common~wa122?ver=13.0.1",
            "address": "119.119.192.194",
            "connectionLimit": 0,
            "dynamicRatio": 1,
            "ephemeral": "false",
            "fqdn": {
                "addressFamily": "ipv4",
                "autopopulate": "disabled",
                "downInterval": 5,
                "interval": "3600",
            },
            "logging": "disabled",
            "monitor": "default",
            "rateLimit": "disabled",
            "ratio": 1,
            "session": "user-enabled",
            "state": "unchecked",
        }

        self.assertEqual(post_icr_session.json(), output)
        self.assertEqual(post_icr_session.json()["name"], "wa122")
        self.assertEqual(post_icr_session.json()["session"], "user-enabled")

    @patch("icontrol.session.iControlRESTSession.patch")
    def test_patch(self, mock_connect_patch):
        full_url = "https://" + self.ip + "/mgmt/tm/ltm/node/~Common~wa122"
        token = "3UCPWZW66ZHOR6BUMVW56F6Q6K"

        mock_connect_patch.return_value.status_code = 200
        mock_connect_patch.return_value = FakeResponsePatch()

        icr_session = iControlRESTSession(
            self.username, self.password, token_to_use=token
        )

        data = {"session": "user-disabled"}

        patch_icr_session = icr_session.patch(full_url, json=data)

        output = {
            "kind": "tm:ltm:node:nodestate",
            "name": "wa122",
            "partition": "Common",
            "fullPath": "/Common/wa122",
            "generation": 2140,
            "selfLink": "https://localhost/mgmt/tm/ltm/node/~Common~wa122?ver=13.0.1",
            "address": "119.119.192.194",
            "connectionLimit": 0,
            "dynamicRatio": 1,
            "ephemeral": "false",
            "fqdn": {
                "addressFamily": "ipv4",
                "autopopulate": "disabled",
                "downInterval": 5,
                "interval": "3600",
            },
            "logging": "disabled",
            "monitor": "default",
            "rateLimit": "disabled",
            "ratio": 1,
            "session": "user-disabled",
            "state": "unchecked",
        }

        self.assertEqual(patch_icr_session.json(), output)
        self.assertEqual(patch_icr_session.json()["name"], "wa122")
        self.assertEqual(patch_icr_session.json()["session"], "user-disabled")

    @patch("icontrol.session.iControlRESTSession.put")
    def test_put(self, mock_connect_put):
        full_url = "https://" + self.ip + "/mgmt/tm/ltm/pool/wa12"
        token = "3UCPWZW66ZHOR6BUMVW56F6Q6K"

        mock_connect_put.return_value.status_code = 200
        mock_connect_put.return_value = FakeResponsePut()

        icr_session = iControlRESTSession(
            self.username, self.password, token_to_use=token
        )

        data = {"members": "wa122:80"}

        put_icr_session = icr_session.put(full_url, json=data)

        output = {
            "kind": "tm:ltm:pool:poolstate",
            "name": "wa12",
            "fullPath": "wa12",
            "generation": 2142,
            "selfLink": "https://localhost/mgmt/tm/ltm/pool/wa12?ver=13.0.1",
            "allowNat": "yes",
            "allowSnat": "yes",
            "ignorePersistedWeight": "disabled",
            "ipTosToClient": "pass-through",
            "ipTosToServer": "pass-through",
            "linkQosToClient": "pass-through",
            "linkQosToServer": "pass-through",
            "loadBalancingMode": "round-robin",
            "minActiveMembers": 0,
            "minUpMembers": 0,
            "minUpMembersAction": "failover",
            "minUpMembersChecking": "disabled",
            "queueDepthLimit": 0,
            "queueOnConnectionLimit": "disabled",
            "queueTimeLimit": 0,
            "reselectTries": 0,
            "serviceDownAction": "none",
            "slowRampTime": 10,
            "membersReference": {
                "link": "https://localhost/mgmt/tm/ltm/pool/~Common~wa12/members?ver=13.0.1",
                "isSubcollection": True,
            },
        }

        self.assertEqual(put_icr_session.json(), output)

    @patch("icontrol.session.iControlRESTSession.delete")
    def test_delete(self, mock_connect_delete):
        full_url = "https://" + self.ip + "/mgmt/tm/ltm/pool/wa12"
        token = "3UCPWZW66ZHOR6BUMVW56F6Q6K"

        mock_connect_delete.return_value.status_code = 200

        icr_session = iControlRESTSession(
            self.username, self.password, token_to_use=token
        )

        delete_icr_session = icr_session.delete(full_url)

        self.assertEqual(delete_icr_session.status_code, 200)

if __name__ == "__main__":
    unittest.main()



import json
import logging
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation
from rest.connector.utils import get_username_password

# create a logger for this module
log = logging.getLogger(__name__)

STD_HEADER = {'Content-type': 'application/json'}

class Implementation(Implementation):
    '''Rest Implementation for DNAC

    Implementation of Rest connection to devices based on pyATS BaseConnection
    for DNAC

    YAML Example
    ------------

        devices:
            dnac1:
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip : "2.3.4.5"
                        port: 443
                        verify: False
                        credentials:
                            rest:
                                username: admin
                                password: cisco123

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/dna.yaml')
        >>> device = testbed.devices['dnac1']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''

    @BaseConnection.locked
    def connect(self, timeout=30):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.
        '''

        if self.connected:
            return

        # Building out Auth request. 
        # prefer host instead of ip address directly
        try:
            host = self.connection_info['host']
        except KeyError:
            host = self.connection_info['ip'].exploded
        
        port = self.connection_info.get('port', 443)
        self.verify = self.connection_info.get('verify', True)

        username, password = get_username_password(self)

        self.base_url = 'https://{host}:{port}'.format(host=host, port=port)

        auth_url  = '{url}/dna/system/api/v1/auth/token'.format(url=self.base_url)
        resp = requests.post(auth_url, headers = STD_HEADER,
                             auth=HTTPBasicAuth(username, password), 
                             verify=self.verify)

        self.token = resp.json()['Token']    # Retrieve the Token from the returned JSONhahhah

        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def get(self, api_url, timeout=30, **kwargs):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            api_url (string): API url string
            timeout: timeout in seconds (default: 30)
            **kwargs: keyword arguments supported in requests.get method
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        full_url = '{url}{api_url}'.format(url=self.base_url,
                                           api_url=api_url)

        log.debug("Sending GET command to '{d}':"\
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        hdr = {'x-auth-token': self.token, 'content-type' : 'application/json'}
        response = requests.get(full_url, headers=hdr,
                                verify=self.verify, timeout=timeout, **kwargs)
        log.info("Output received:\n{response}".format(response=response))

        return response

    @BaseConnection.locked
    def put(self, api_url, timeout=30, **kwargs):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            api_url (string): API url string
            timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        full_url = '{url}{api_url}'.format(url=self.base_url,
                                           api_url=api_url)

        log.debug("Sending PUT command to '{d}':"\
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        hdr = {'x-auth-token': self.token, 'content-type' : 'application/json'}
        response = requests.put(full_url, headers=hdr,
                                verify=self.verify, timeout=timeout, **kwargs)
        log.info("Output received:\n{response}".format(response=response.text))

        return response

    @BaseConnection.locked
    def post(self, api_url, timeout=30, **kwargs):
        '''GET REST Command to POST information to the device

        Arguments
        ---------

            api_url (string): API url string
            timeout: timeout in seconds (default: 30)
            **kwargs are the same keyword arguments for the requests lib
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        full_url = '{url}{api_url}'.format(url=self.base_url,
                                           api_url=api_url)

        log.info("Sending POST command to '{d}':"\
                 "\nDN: {furl}".format(d=self.device.name, furl=full_url))

        hdr = {'x-auth-token': self.token, 'content-type' : 'application/json'}
        response = requests.post(full_url, headers=hdr,
                                verify=self.verify, timeout=timeout, **kwargs)
        log.info("Output received:\n{response}".format(response=response))

        return response



import logging
import requests

from requests.exceptions import RequestException


from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation
from rest.connector.utils import get_username_password

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(Implementation):
    '''Rest Implementation for VIRL

    Implementation of Rest connection to devices based on pyATS BaseConnection
    for VIRL

    YAML Example
    ------------

        devices:
            virl:
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip : "192.168.1.1"
                        port: "19399"
                        protocol: http
                        credentials:
                            default:
                                username: admin
                                password: cisco123

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
        >>> device = testbed.devices['virl']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''

    @BaseConnection.locked
    def connect(self, timeout=30, port="19399", protocol='http'):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value
            port (str): Port number. Default to 19399
            protocol (str): http or https

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        There is no return from this method. If something goes wrong, an
        exception will be raised.


        YAML Example
        ------------

            devices:
                virl:
                    connections:
                        rest:
                            class: rest.connector.Rest
                            ip : "192.168.1.1"
                            port: "80"
                            protocol : http
                            credentials:
                                default:
                                    username: admin
                                    password: cisco123

        Code Example
        ------------

            >>> from pyats.topology import loader
            >>> testbed = loader.load('/users/xxx/xxx/testbed.yaml')
            >>> device = testbed.devices['virl']
            >>> device.connect(alias='rest', via='rest')
        '''

        if self.connected:
            return

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    '`unicon` is not installed for `sshtunnel`. Please install by `pip install unicon`.'
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            ip = self.connection_info['ip'].exploded
            port = self.connection_info.get('port', '19399')

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.url = '{protocol}://{ip}:{port}'.format(protocol=protocol,
                                                          ip=ip,
                                                          port=port)

        self.username, self.password = get_username_password(self)
        self.headers = {"Content-Type": "text/xml;charset=UTF-8"}

        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))

        self.session = requests.Session()

        # Connect to the device via requests
        if protocol == 'https':
            response = self.session.get(self.url + '/roster/rest/test',
                                        auth=(self.username, self.password),
                                        timeout=timeout,
                                        headers=self.headers,
                                        verify=False)
        else:
            response = self.session.get(self.url + '/roster/rest/test',
                                        auth=(self.username, self.password),
                                        timeout=timeout,
                                        headers=self.headers)
        log.info(response)

        # Make sure it returned requests.codes.ok
        if response.status_code != requests.codes.ok:
            # Something bad happened
            raise RequestException("Connection to '{ip}' has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{ok}'"\
                                        .format(ip=ip, c=response.status_code,
                                                ok=requests.codes.ok))
        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

    @BaseConnection.locked
    def disconnect(self):
        '''disconnect the device for this particular alias'''

        log.info("Disconnecting from '{d}' with "
                 "alias '{a}'".format(d=self.device.name, a=self.alias))
        try:
            self.session.close()
        finally:
            self._is_connected = False
        log.info("Disconnected successfully from "
                 "'{d}'".format(d=self.device.name))

    def isconnected(func):
        '''Decorator to make sure session to device is active

           There is limitation on the amount of time the session cab be active
           on the VIRL. However, there are no way to verify if
           session is still active unless sending a command. So, its just
           faster to reconnect every time.
         '''
        def decorated(self, *args, **kwargs):
            # Check if connected
            try:
                self.disconnect()

                if 'timeout' in kwargs:
                    self.connect(timeout=kwargs['timeout'])
                else:
                    self.connect()

            finally:
                ret = func(self, *args, **kwargs)
            return ret
        return decorated

    @BaseConnection.locked
    @isconnected
    def get(self, url, \
            expected_status_code=requests.codes.ok, timeout=30, **kwargs):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            url (string): REST API url
            expected_status_code (int): Expected result
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = "{f}{url}"\
            .format(f=self.url, url=url)

        log.debug("Sending GET command to '{d}':"\
                 "\nurl: {url}".format(d=self.device.name, url=full_url))

        response = self.session.get(full_url,
                                    auth=(self.username, self.password),
                                    timeout=timeout,
                                    headers=self.headers,
                                    **kwargs)

        try:
            output = response.json()
        except Exception:
            output = response.text

        log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("Sending '{furl} to '{d} has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{e}'"
                                   "'{e}'".format(furl=full_url,
                                                  d=self.device.name,
                                                  c=response.status_code,
                                                  e=expected_status_code))
        return output

    @BaseConnection.locked
    @isconnected
    def post(self, url, payload, expected_status_code=requests.codes.ok,
             timeout=30, **kwargs):
        '''POST REST Command to configure information from the device

        Arguments
        ---------

            url (string): API url
            payload (dict): Dictionary containing the information to send via
                            the post
            expected_status_code (int): Expected result
            timeout (int): Maximum time
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        # Deal with the dn
        full_url = '{f}{url}'.format(f=self.url, url=url)

        log.debug("Sending POST command to '{d}':"\
                 "\nDN: {furl}\nPayload:{p}".format(d=self.device.name,
                                                    furl=full_url,
                                                    p=payload))

        # Send to the device
        response = self.session.post(full_url,
                                     payload,
                                     auth=(self.username, self.password),
                                     timeout=timeout,
                                     headers=self.headers,
                                     **kwargs)
        try:
            output = response.json()
        except Exception:
            output = response.text

        log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}', got:\n {msg}"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code,
                                           msg=response.text))
        return output

    @BaseConnection.locked
    @isconnected
    def delete(self, url, expected_status_code=requests.codes.ok, timeout=30, **kwargs):
        '''DELETE REST Command to delete information from the device

        Arguments
        ---------

            url (string): API url
            expected_status_code (int): Expected result
            timeout (int): Maximum time
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = '{f}{url}'.format(f=self.url, url=url)

        log.debug("Sending DELETE command to '{d}':"\
                 "\nurl: {url}".format(d=self.device.name, url=full_url))

        # Send to the device
        response = self.session.delete(full_url,
                                       auth=(self.username, self.password),
                                       headers=self.headers,
                                       timeout=timeout,
                                       **kwargs)

        try:
            output = response.json()
        except Exception:
            output = response.text

        log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code != expected_status_code:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code "
                                   "'{e}' for '{d}'"\
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_code))
        return output

    @BaseConnection.locked
    def put(self, url, timeout=30, **kwargs):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------

            url (string): API url
            timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = '{f}{url}'.format(f=self.url,
                                           url=url)

        log.debug("Sending PUT command to '{d}':"\
                 "\nurl: {url}".format(d=self.device.name, url=full_url))

        response = requests.put(full_url, auth=(self.username, self.password),
                                headers=self.headers, timeout=timeout, **kwargs)

        try:
            output = response.json()
        except Exception:
            output = response.text

        log.info("Output received:\n{output}".format(output=output))

        return output



#!/bin/env python
""" Unit tests for viptela/vManage rest.connector """

import os
import unittest
from requests.models import Response
from unittest.mock import patch, MagicMock
from requests.exceptions import RequestException

from pyats.topology import loader

from rest.connector import Rest
HERE = os.path.dirname(__file__)


class test_rest_connector(unittest.TestCase):

    def setUp(self):
        self.testbed = loader.load(os.path.join(HERE, 'testbed.yaml'))
        self.device = self.testbed.devices['vmanage']

    def test_init(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.device, self.device)

        with self.assertRaises(NotImplementedError):
            self.assertRaises(connection.execute())
        with self.assertRaises(NotImplementedError):
            self.assertRaises(connection.configure())

    def test_connection(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 200
            req().post.return_value = resp
            req().get.return_value = resp
            connection.connect()
            self.assertEqual(connection.connected, True)
            connection.connect()
            self.assertEqual(connection.connected, True)

        # Now disconnect
        with patch('requests.session') as req:
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_post_not_connected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        with self.assertRaises(Exception):
            connection.post(mount_point='temp',
                            payload={'payload': 'something'})

    def test_connection_wrong_code(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 404
            req().post.return_value = resp
            req().get.return_value = resp

            with self.assertRaises(RequestException):
                connection.connect()

        self.assertEqual(connection.connected, False)

        # Now disconnect
        with patch('requests.session') as req:
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_post_connected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 200
            req().post.return_value = resp
            req().get.return_value = resp
            connection.connect()
            resp.json = MagicMock(return_value={'imdata': []})
            connection.post(mount_point='temp',
                            payload={'payload': 'something'})
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_get_not_connected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        with self.assertRaises(Exception):
            connection.get(mount_point='temp')

    def test_get_connected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 200
            req().post.return_value = resp
            req().get.return_value = resp
            connection.connect()
            resp.json = MagicMock(return_value={'imdata': []})
            connection.get(mount_point='temp')
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_get_connected_change_expected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 200
            resp2 = Response()
            resp2.status_code = 300
            req().get.side_effect = [resp, resp2]
            req().post.side_effect = [resp, resp, resp2]

            connection.connect()
            resp.json = MagicMock(return_value={'imdata': []})
            resp2.json = MagicMock(return_value={'imdata': []})

            connection.get(mount_point='temp')
            self.assertEqual(connection.connected, True)
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_delete_not_connected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        with self.assertRaises(Exception):
            connection.delete(mount_point='temp')

    def test_delete_connected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 200
            req().post.return_value = resp
            req().get.return_value = resp
            req().delete.return_value = resp
            connection.connect()
            resp.json = MagicMock(return_value={'imdata': []})
            connection.delete(mount_point='temp')
            connection.disconnect()
        self.assertEqual(connection.connected, False)

    def test_delete_connected_change_expected(self):
        connection = Rest(device=self.device, alias='rest', via='rest')
        self.assertEqual(connection.connected, False)

        with patch('requests.session') as req:
            resp = Response()
            resp.status_code = 200
            resp2 = Response()
            resp2.status_code = 300
            req().delete.return_value = resp2
            req().post.side_effect = [resp, resp, resp2]
            req().get.return_value = resp
            connection.connect()
            resp.json = MagicMock(return_value={'imdata': []})
            resp2.json = MagicMock(return_value={'imdata': []})
            connection.delete(mount_point='temp')
            self.assertEqual(connection.connected, True)
            connection.disconnect()

        self.assertEqual(connection.connected, False)


if __name__ == '__main__':
    unittest.main()



import json
import logging
import re
import urllib.request
import requests
from dicttoxml import dicttoxml
from requests.exceptions import RequestException

from pyats.connections import BaseConnection
from rest.connector.implementation import Implementation as RestImplementation
from rest.connector.utils import get_username_password

# create a logger for this module
log = logging.getLogger(__name__)


class Implementation(RestImplementation):
    '''Rest Implementation for IOS-XE

    Implementation of Rest connection to elasitc Wireless LAN Controller running IOS-XE

    YAML Example
    ------------

        devices:
            eWLC:
                os: iosxe
                connections:
                    rest:
                        class: rest.connector.Rest
                        ip: 127.0.0.1
                        port: "443"
                        protocol: https
                        credentials:
                            rest:
                                username: admin
                                password: admin
                custom:
                    abstraction:
                        order: [os]

    Code Example
    ------------

        >>> from pyats.topology import loader
        >>> testbed = loader.load('topology.yaml')
        >>> device = testbed.devices['eWLC']
        >>> device.connect(alias='rest', via='rest')
        >>> device.rest.connected
        True
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'proxies' not in kwargs:
            self.proxies = urllib.request.getproxies()

    @BaseConnection.locked
    def connect(self,
                timeout=30,
                default_content_type='json',
                verbose=False,
                port="443",
                protocol='https'):
        '''connect to the device via REST

        Arguments
        ---------

            timeout (int): Timeout value
            default_content_type: Default for content type, json or xml
            proxies: Specify the proxy to use for connection as seen below.
                    {'http': 'http://proxy.esl.cisco.com:80/',
                    'ftp': 'http://proxy.esl.cisco.com:80/',
                    'https': 'http://proxy.esl.cisco.com:80/',
                    'no': '.cisco.com'}

        Raises
        ------

        Exception
        ---------

            If the connection did not go well

        Note
        ----

        Connecting via RESTCONF does not require contacting the device.
        This does nothing

        '''
        if self.connected:
            return

        log.debug("Content type: %s" % default_content_type)
        log.debug("Timeout: %s" % timeout)
        self.content_type = default_content_type

        # support sshtunnel
        if 'sshtunnel' in self.connection_info:
            try:
                from unicon.sshutils import sshtunnel
            except ImportError:
                raise ImportError(
                    '`unicon` is not installed for `sshtunnel`. Please install by `pip install unicon`.'
                )
            try:
                tunnel_port = sshtunnel.auto_tunnel_add(self.device, self.via)
                if tunnel_port:
                    ip = self.device.connections[self.via].sshtunnel.tunnel_ip
                    port = tunnel_port
            except AttributeError as e:
                raise AttributeError(
                    "Cannot add ssh tunnel. Connection %s may not have ip/host or port.\n%s"
                    % (self.via, e))
        else:
            ip = self.connection_info.ip.exploded
            port = self.connection_info.get('port', port)

        if 'protocol' in self.connection_info:
            protocol = self.connection_info['protocol']

        self.base_url = '{protocol}://{ip}:{port}'.format(protocol=protocol,
                                                          ip=ip,
                                                          port=port)
        self.login_url = '{f}/'.format(f=self.base_url)
        log.info("Connecting to '{d}' with alias "
                 "'{a}'".format(d=self.device.name, a=self.alias))
        username, password = get_username_password(self)
        self.session = requests.Session()
        self.session.auth = (username, password)
        # Connect to the device via requests
        response = self.session.get(
            self.login_url, proxies=self.proxies, timeout=timeout, verify=False)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
                                                           r=response.reason, h=response.headers))
        if verbose:
            log.info("Response text:\n%s" % output)

        # Make sure it returned requests.codes.ok
        if response.status_code != requests.codes.ok:
            # Something bad happened
            raise RequestException("Connection to '{ip}:{port}' has returned the "
                                   "following code '{c}', instead of the "
                                   "expected status code '{ok}'"
                                   .format(ip=ip, port=port, c=response.status_code,
                                           ok=requests.codes.ok))
        self._is_connected = True
        log.info("Connected successfully to '{d}'".format(d=self.device.name))

        return response

    @BaseConnection.locked
    def disconnect(self):
        """
            Does not make sense to disconnect from a device.
        """
        self._is_connected = False
        return

    @BaseConnection.locked
    def get(self, api_url, content_type=None, headers=None,
            expected_status_codes=(
                requests.codes.no_content,
                requests.codes.ok
            ),
            timeout=30,
            verbose=False):
        '''GET REST Command to retrieve information from the device

        Arguments
        ---------
        api_url: API url string
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''
        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))
        if content_type is None:
            content_type = self.content_type

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        header = 'application/yang-data+{fmt}'

        if content_type.lower() == 'json':
            accept_header = header.format(fmt='json')
        elif content_type.lower() == 'xml':
            accept_header = header.format(fmt='xml')
        else:
            accept_header = content_type

        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending GET command to '{d}': "
                 "{u}".format(d=self.device.name, u=full_url))
        log.debug("Request headers:{headers}".format(
            headers=self.session.headers))

        response = self.session.get(full_url, proxies=self.proxies, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
                                                           r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response

    @BaseConnection.locked
    def post(self, api_url, payload='', content_type=None, headers=None,
             expected_status_codes=(
                 requests.codes.created,
                 requests.codes.no_content,
                 requests.codes.ok
             ),
             timeout=30,
             verbose=False):
        '''POST REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        payload: payload to sent, can be string or dict
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        request_payload = payload
        if isinstance(payload, dict):
            assert content_type is not None, 'content_type parameter required when passing dict'
            if content_type == 'json':
                request_payload = json.dumps(payload)
            elif content_type == 'xml':
                request_payload = dicttoxml(payload, attr_type=False)

        if content_type is None:
            if re.match("<", payload.lstrip()) is not None:
                content_type = 'xml'
            else:
                content_type = 'json'

        content_type_header = 'application/yang-data+{fmt}'
        accept_header = 'application/yang-data+{fmt}'

        if content_type.lower() == 'json':
            content_type_header = content_type_header.format(fmt='json')
            accept_header = accept_header.format(fmt='json')
        elif content_type.lower() == 'xml':
            content_type_header = content_type_header.format(fmt='xml')
            accept_header = accept_header.format(fmt='xml')
        else:
            content_type_header = content_type
            accept_header = content_type

        self.session.headers.update({'Content-type': content_type_header})
        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending POST command to '{d}': {u}"
                 .format(d=self.device.name, u=full_url))
        log.debug("Request headers: {h}\nPayload: {p}"
                  .format(h=self.session.headers, p=request_payload))
        if verbose:
            log.info('Request payload:\n{payload}'.format(
                payload=request_payload))

        # Send to the device
        response = self.session.post(
            full_url, request_payload, proxies=self.proxies, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
                                                           r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response

    @BaseConnection.locked
    def patch(self, api_url, payload, content_type=None, headers=None,
              expected_status_codes=(
                  requests.codes.created,
                  requests.codes.no_content,
                  requests.codes.ok
              ),
              timeout=30,
              verbose=False):
        '''PATCH REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        payload: payload to sent, can be string or dict
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        request_payload = payload
        if isinstance(payload, dict):
            assert content_type is not None, 'content_type parameter required when passing dict'
            if content_type == 'json':
                request_payload = json.dumps(payload)
            elif content_type == 'xml':
                request_payload = dicttoxml(payload, attr_type=False)

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        if content_type is None:
            if re.match("<", payload.lstrip()) is not None:
                content_type = 'xml'
            else:
                content_type = 'json'

        content_type_header = 'application/yang-data+{fmt}'
        accept_header = 'application/yang-data+{fmt}'

        if content_type.lower() == 'json':
            content_type_header = content_type_header.format(fmt='json')
            accept_header = accept_header.format(fmt='json')
        elif content_type.lower() == 'xml':
            content_type_header = content_type_header.format(fmt='xml')
            accept_header = accept_header.format(fmt='xml')
        else:
            content_type_header = content_type
            accept_header = content_type

        self.session.headers.update({'Content-type': content_type_header})
        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending PATCH command to '{d}': {u}".format(
            d=self.device.name, u=full_url))
        log.debug("Request headers: {h}\nPayload:{p}".format(h=self.session.headers,
                                                             p=request_payload))
        if verbose:
            log.info('Request payload:\n{payload}'.format(
                payload=request_payload))

        # Send to the device
        response = self.session.patch(
            full_url, request_payload, proxies=self.proxies, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
                                                           r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response

    @BaseConnection.locked
    def put(self, api_url, payload, content_type=None, headers=None,
            expected_status_codes=(
                requests.codes.created,
                requests.codes.no_content,
                requests.codes.ok
            ),
            timeout=30,
            verbose=False):
        '''PUT REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        payload: payload to sent, can be string or dict
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        request_payload = payload
        if isinstance(payload, dict):
            assert content_type != None, 'content_type parameter required when passing dict'
            if content_type == 'json':
                request_payload = json.dumps(payload)
            elif content_type == 'xml':
                request_payload = dicttoxml(payload, attr_type=False)

        if content_type is None:
            if re.match("<", payload.lstrip()) is not None:
                content_type = 'xml'
            else:
                content_type = 'json'

        content_type_header = 'application/yang-data+{fmt}'
        accept_header = 'application/yang-data+{fmt}'

        if content_type.lower() == 'json':
            content_type_header = content_type_header.format(fmt='json')
            accept_header = accept_header.format(fmt='json')
        elif content_type.lower() == 'xml':
            content_type_header = content_type_header.format(fmt='xml')
            accept_header = accept_header.format(fmt='xml')
        else:
            content_type_header = content_type
            accept_header = content_type

        self.session.headers.update({'Content-type': content_type_header})
        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending PUT command to '{d}': {u}".format(
            d=self.device.name, u=full_url))
        log.debug("Request headers: {h}\nPayload:{p}".format(h=self.session.headers,
                                                             p=request_payload))
        if verbose:
            log.info('Request payload:\n{payload}'.format(
                payload=request_payload))

        # Send to the device
        response = self.session.put(full_url, request_payload, proxies=self.proxies, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
                                                           r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response

    @BaseConnection.locked
    def delete(self, api_url, content_type=None, headers=None,
               expected_status_codes=(
                   requests.codes.created,
                   requests.codes.no_content,
                   requests.codes.ok
               ),
               timeout=30,
               verbose=False):
        '''DELETE REST Command to configure information from the device

        Arguments
        ---------
        api_url: API url string
        content_type: expected content type to be returned (xml or json)
        headers: dictionary of HTTP headers (optional)
        expected_status_codes: list of expected result codes (integers)
        timeout: timeout in seconds (default: 30)
        '''

        if not self.connected:
            raise Exception("'{d}' is not connected for "
                            "alias '{a}'".format(d=self.device.name,
                                                 a=self.alias))

        if content_type is None:
            content_type = self.content_type

        full_url = '{b}{a}'.format(b=self.base_url, a=api_url)

        if content_type.lower() == 'json':
            accept_header = 'application/yang-data+json'
        elif content_type.lower() == 'xml':
            accept_header = 'application/yang-data+xml'
        else:
            accept_header = content_type

        self.session.headers.update({'Accept': accept_header})
        if headers is not None:
            self.session.headers.update(headers)

        log.debug("Sending DELETE command to '{d}': "
                 "{u}".format(d=self.device.name, u=full_url))
        log.debug("Request headers:{headers}".format(
            headers=self.session.headers))

        response = self.session.delete(full_url, proxies=self.proxies, timeout=timeout)
        output = response.text
        log.debug("Response: {c} {r}, headers: {h}".format(c=response.status_code,
                                                           r=response.reason, h=response.headers))
        if verbose:
            log.info("Output received:\n{output}".format(output=output))

        # Make sure it returned requests.codes.ok
        if response.status_code not in expected_status_codes:
            # Something bad happened
            raise RequestException("'{c}' result code has been returned "
                                   "instead of the expected status code(s) "
                                   "'{e}' for '{d}'\n{t}"
                                   .format(d=self.device.name,
                                           c=response.status_code,
                                           e=expected_status_codes,
                                           t=response.text))
        return response
